#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
import datetime
import warnings
from collections.abc import Iterator
from typing import Any, Optional, Union

from elementpath.aliases import NamespacesType, InputType, ParserClassType, \
    RootArgType, ItemArgType
from elementpath.xpath_context import XPathContext
from elementpath.xpath2 import XPath2Parser
from elementpath.datatypes import Timezone
from elementpath.schema_proxy import AbstractSchemaProxy


def select(root: Optional[RootArgType],
           path: str,
           namespaces: Optional[NamespacesType] = None,
           parser: Optional[ParserClassType] = None,
           uri: Optional[str] = None,
           fragment: Optional[bool] = None,
           item: Optional[ItemArgType] = None,
           position: int = 1,
           size: int = 1,
           axis: Optional[str] = None,
           schema: Optional[AbstractSchemaProxy] = None,
           variables: Optional[dict[str, InputType[ItemArgType]]] = None,
           current_dt: Optional[datetime.datetime] = None,
           timezone: Optional[Union[str, Timezone]] = None,
           **kwargs: Any) -> Any:
    """
    XPath selector function that apply a *path* expression on *root* Element.

    :param root: the root of the XML document, usually an ElementTree instance or an \
    Element. A schema or a schema element can also be provided, or an already built \
    node tree. You can also provide `None`, in which case no XML root node is set in \
    the dynamic context, and you have to provide the keyword argument *item*.
    :param path: the XPath expression.
    :param namespaces: a dictionary with mapping from namespace prefixes into URIs.
    :param parser: the parser class to use, that is :class:`XPath2Parser` for default.
    :param uri: an optional URI associated with the root element or the document.
    :param fragment: if `True` is provided the root is considered a fragment. In this \
    case if `root` is an ElementTree instance skips it and use the root Element. If \
    `False` is provided creates a dummy document when the root is an Element instance. \
    In this case the dummy document value is not included in results. For default the \
    root node kind is preserved.
    :param item: the context item. A `None` value means that the context is positioned on \
    the document node.
    :param position: the current position of the node within the input sequence.
    :param size: the number of items in the input sequence.
    :param axis: the active axis. Used to choose when apply the default axis ('child' axis).
    :param schema: an optional schema proxy instance for applying XSD type annotations \
    on element and attribute nodes.
    :param variables: dictionary of context variables that maps a QName to a value.
    :param current_dt: current dateTime of the implementation, including explicit timezone.
    :param timezone: implicit timezone to be used when a date, time, or dateTime value does \
    not have a timezone.
    :param kwargs: other optional parameters for the parser instance.
    :return: a list with XPath nodes or a basic type for expressions based \
    on a function or literal.
    """
    if parser is None:
        parser = XPath2Parser
    if schema is not None and parser.version > '1.0':
        kwargs['schema'] = schema

    root_token = parser(namespaces, **kwargs).parse(path)
    context = XPathContext(root, namespaces, uri, fragment, item, position, size,
                           axis, schema, variables, current_dt, timezone)
    return root_token.get_results(context)


def iter_select(root: Optional[RootArgType],
                path: str,
                namespaces: Optional[NamespacesType] = None,
                parser: Optional[ParserClassType] = None,
                uri: Optional[str] = None,
                fragment: Optional[bool] = None,
                item: Optional[ItemArgType] = None,
                position: int = 1,
                size: int = 1,
                axis: Optional[str] = None,
                schema: Optional[AbstractSchemaProxy] = None,
                variables: Optional[dict[str, InputType[ItemArgType]]] = None,
                current_dt: Optional[datetime.datetime] = None,
                timezone: Optional[Union[str, Timezone]] = None,
                **kwargs: Any) -> Iterator[Any]:
    """
    A function that creates an XPath selector generator for apply a *path* expression
    on *root* Element.

    :param root: the root of the XML document, usually an ElementTree instance or an \
    Element. A schema or a schema element can also be provided, or an already built \
    node tree. You can also provide `None`, in which case no XML root node is set in \
    the dynamic context, and you have to provide the keyword argument *item*.
    :param path: the XPath expression.
    :param namespaces: a dictionary with mapping from namespace prefixes into URIs.
    :param parser: the parser class to use, that is :class:`XPath2Parser` for default.
    :param uri: an optional URI associated with the root element or the document.
    :param fragment: if `True` is provided the root is considered a fragment. In this \
    case if `root` is an ElementTree instance skips it and use the root Element. If \
    `False` is provided creates a dummy document when the root is an Element instance. \
    In this case the dummy document value is not included in results. For default the \
    root node kind is preserved.
    :param item: the context item. A `None` value means that the context is positioned on \
    the document node.
    :param position: the current position of the node within the input sequence.
    :param size: the number of items in the input sequence.
    :param axis: the active axis. Used to choose when apply the default axis ('child' axis).
    :param schema: an optional schema proxy instance for applying XSD type annotations \
    on element and attribute nodes.
    :param variables: dictionary of context variables that maps a QName to a value.
    :param current_dt: current dateTime of the implementation, including explicit timezone.
    :param timezone: implicit timezone to be used when a date, time, or dateTime value does \
    not have a timezone.
    :param kwargs: other optional parameters for the parser instance.
    :return: a generator of the XPath expression results.
    """
    if parser is None:
        parser = XPath2Parser
    if schema is not None and parser.version > '1.0':
        kwargs['schema'] = schema

    root_token = parser(namespaces, **kwargs).parse(path)
    context = XPathContext(root, namespaces, uri, fragment, item, position, size,
                           axis, schema, variables, current_dt, timezone)
    return root_token.select_results(context)


class Selector(object):
    """
    XPath selector class. Create an instance of this class if you want to apply an XPath
    selector to several target data.

    :param path: the XPath expression.
    :param namespaces: a dictionary with mapping from namespace prefixes into URIs.
    :param parser: the parser class to use, that is :class:`XPath2Parser` for default.
    :param kwargs: other optional parameters for the XPath parser instance.

    :ivar path: the XPath expression.
    :vartype path: str
    :ivar parser: the parser instance.
    :vartype parser: XPath1Parser or XPath2Parser
    :ivar root_token: the root of tokens tree compiled from path.
    :vartype root_token: XPathToken
    """
    def __init__(self, path: str,
                 namespaces: Optional[NamespacesType] = None,
                 parser: Optional[ParserClassType] = None,
                 **kwargs: Any) -> None:

        if 'variables' in kwargs:
            msg = ("Argument 'variables' here is deprecated and will be"
                   "be removed in the next major release. Provide this "
                   "argument later to Selector.select/iter_select.")
            warnings.warn(DeprecationWarning(msg))

        self._variables = kwargs.pop('variables', None)
        self.parser = (parser or XPath2Parser)(namespaces, **kwargs)
        self.path = path
        self.root_token = self.parser.parse(path)

    def __repr__(self) -> str:
        return '%s(path=%r, parser=%s)' % (
            self.__class__.__name__, self.path, self.parser.__class__.__name__
        )

    @property
    def namespaces(self) -> dict[str, str]:
        """A dictionary with mapping from namespace prefixes into URIs."""
        return self.parser.namespaces

    def select(self, root: Optional[RootArgType], **kwargs: Any) -> Any:
        """
        Applies the instance's XPath expression on *root* Element.

        :param root: the root of the XML document, usually an ElementTree instance \
        or an Element.
        :param kwargs: other optional parameters for the XPath dynamic context.
        :return: a list with XPath nodes or a basic type for expressions based on \
        a function or literal.
        """
        if 'schema' not in kwargs:
            kwargs['schema'] = self.parser.schema
        if 'variables' not in kwargs and self._variables:
            kwargs['variables'] = self._variables

        context = XPathContext(root, **kwargs)
        return self.root_token.get_results(context)

    def iter_select(self, root: Optional[RootArgType], **kwargs: Any) -> Iterator[Any]:
        """
        Creates an XPath selector generator for apply the instance's XPath expression
        on *root* Element.

        :param root: the root of the XML document, usually an ElementTree instance \
        or an Element.
        :param kwargs: other optional parameters for the XPath dynamic context.
        :return: a generator of the XPath expression results.
        """
        if 'schema' not in kwargs:
            kwargs['schema'] = self.parser.schema
        if 'variables' not in kwargs and self._variables:
            kwargs['variables'] = self._variables

        context = XPathContext(root, **kwargs)
        return self.root_token.select_results(context)



#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
"""A collection of additional and special token classes."""
import math
from collections.abc import Iterator
from decimal import Decimal
from typing import Literal, cast, NoReturn

import elementpath.aliases as ta

from elementpath.namespaces import XSD_NAMESPACE, XPATH_FUNCTIONS_NAMESPACE, XMLNS_NAMESPACE
from elementpath.namespaces import get_expanded_name
from elementpath.datatypes import AnyAtomicType, AnyURI, UntypedAtomic, ArithmeticProxy, \
    YearMonthDuration, DayTimeDuration, Duration, AbstractDateTime
from elementpath.helpers import collapse_white_spaces
from elementpath.sequences import xlist
from elementpath.xpath_nodes import AttributeNode, ElementNode
from elementpath.xpath_context import XPathSchemaContext
from elementpath.decoder import get_atomic_sequence

from .base import XPathToken


class ValueToken(XPathToken):
    """
    A dummy token for encapsulating a value.
    """
    symbol = '(value)'
    value: AnyAtomicType

    @property
    def source(self) -> str:
        return str(self.value)

    def evaluate(self, context: ta.ContextType = None) -> AnyAtomicType:
        return self.value

    def select(self, context: ta.ContextType = None) -> Iterator[AnyAtomicType]:
        if isinstance(self.value, list):
            yield from self.value
        else:
            yield self.value


###
# Token classes for names
class ProxyToken(XPathToken):
    """
    A token class for resolving collisions between other tokens that have
    the same symbol but are in different namespaces. It also resolves
    collisions of functions with names.
    """
    symbol = '(proxy)'

    def nud(self) -> 'XPathToken':
        if self.parser.next_token.symbol not in ('(', '#'):
            # Not a function call or reference, returns a name.
            return self.as_name()

        lookup_name = f'{{{self.namespace or XPATH_FUNCTIONS_NAMESPACE}}}{self.value}'
        try:
            token = self.parser.symbol_table[lookup_name](self.parser)
        except KeyError:
            if self.namespace == XSD_NAMESPACE:
                msg = f'unknown constructor function {self.symbol!r}'
            else:
                msg = f'unknown function {self.symbol!r}'
            raise self.error('XPST0017', msg) from None
        else:
            if self.parser.next_token.symbol == '#':
                return token

            res = token.nud()
            return res


###
# Name related tokens for matching elements and attributes
class NameToken(XPathToken):
    """
    The special '(name)' token for matching attributes or element nodes.
    For XPath its value is an unprefixed name.
    """
    symbol = lookup_name = '(name)'
    label = 'name'
    bp = 10
    value: str

    def __str__(self) -> str:
        return f'{self.value!r} name'

    def nud(self) -> XPathToken:
        if self.parser.next_token.symbol == '::':
            msg = "axis '%s::' not found" % self.value
            if self.parser.compatibility_mode:
                raise self.error('XPST0010', msg)
            raise self.error('XPST0003', msg)
        elif self.parser.next_token.symbol == '(':
            if self.parser.version >= '2.0':
                pass  # XP30+ has led() for '(' operator that can check this
            elif self.namespace == XSD_NAMESPACE:
                raise self.error('XPST0017', 'unknown constructor function {!r}'.format(self.value))
            elif self.namespace or self.value not in self.parser.RESERVED_FUNCTION_NAMES:
                raise self.error('XPST0017', 'unknown function {!r}'.format(self.value))
            else:
                msg = f"{self.value!r} is not allowed as function name"
                raise self.error('XPST0003', msg)

        return self

    def evaluate(self, context: ta.ContextType = None) -> list[AttributeNode | ElementNode]:
        return xlist(self.select(context))

    def select(self, context: ta.ContextType = None) -> Iterator[AttributeNode | ElementNode]:
        if context is None:
            raise self.missing_context()
        yield from context.iter_matching_nodes(self.value, self.parser.default_namespace)


class PrefixedNameToken(XPathToken):
    """Colon symbol for expressing prefixed qualified names or wildcards."""
    symbol = lookup_name = ':'
    lbp = 95
    rbp = 95
    value: str

    def __init__(self, parser: ta.XPathParserType, value: Literal[':'] = ':') -> None:
        super().__init__(parser, value)

        # Change bind powers if it cannot be a namespace related token
        if self.is_spaced():
            self.lbp = self.rbp = 0
        elif self.parser.token.symbol not in ('*', '(name)', 'array'):
            self.lbp = self.rbp = 0

    def __str__(self) -> str:
        if len(self) < 2:
            return 'unparsed prefixed reference'
        elif self[1].label.endswith('function'):
            return f"{self.value!r} {self[1].label}"
        elif '*' in self.value:
            return f"{self.value!r} prefixed wildcard"
        else:
            return f"{self.value!r} prefixed name"

    @property
    def source(self) -> str:
        return ':'.join(tk.source for tk in self) + self.occurrence

    def led(self, left: XPathToken) -> XPathToken:
        version = self.parser.version
        if self.is_spaced():
            if version <= '3.0':
                raise self.wrong_syntax("a QName cannot contains spaces before or after ':'")
            return left

        if version == '1.0':
            left.expected('(name)')
        elif version <= '3.0':
            left.expected('(name)', '*')
        elif left.symbol not in ('(name)', '*'):
            return left

        if not self.parser.next_token.label.endswith('function'):
            self.parser.expected_next('(name)', '*')

        if isinstance(left, NameToken):
            try:
                namespace = self.parser.namespaces[left.value]
            except KeyError:
                self.parser.advance()  # Assure there isn't a following incomplete comment
                self[:] = left, self.parser.token
                msg = "prefix {!r} is not declared".format(left.value)
                # raise self.error('FONS0004', msg) from None  FIXME?? XP30+??
                raise self.error('XPST0081', msg) from None
            else:
                self.parser.next_token.bind_namespace(namespace)
        elif self.parser.next_token.symbol != '(name)':
            raise self.wrong_syntax()
        else:
            self.parser.next_token.bind_namespace('*')

        self[:] = left, self.parser.expression(95)

        self.name = self[1].name
        if self[1].label.endswith('function'):
            self.value = f'{self[0].value}:{self[1].symbol}'
        else:
            self.value = f'{self[0].value}:{self[1].value}'
        return self

    def evaluate(self, context: ta.ContextType = None) -> ta.ValueType:
        if self[1].label.endswith('function'):
            return self[1].evaluate(context)
        return xlist(self.select(context))

    def select(self, context: ta.ContextType = None) -> Iterator[ta.ItemType]:
        if self[1].label.endswith('function'):
            value = self[1].evaluate(context)
            if isinstance(value, list):
                yield from value
            elif value is not None:
                yield value
            return

        if context is None:
            raise self.missing_context()

        yield from context.iter_matching_nodes(self.name)


class BracedNameToken(XPathToken):
    """Braced expanded name for expressing namespace related names."""

    symbol = lookup_name = '{'
    label = 'expanded name'
    bp = lbp = rbp = 95
    value: str

    def nud(self) -> XPathToken:
        if self.parser.strict and self.symbol == '{':
            raise self.wrong_syntax("not allowed symbol if parser has strict=True")

        self.parser.next_token.unexpected('{')
        if self.parser.next_token.symbol == '}':
            namespace = ''
        else:
            value = self.parser.next_token.value
            assert isinstance(value, str)
            namespace = value + self.parser.advance_until('}')
            namespace = collapse_white_spaces(namespace)

        try:
            AnyURI(namespace)
        except ValueError as err:
            msg = f"invalid URI in an EQName: {str(err)}"
            raise self.error('XQST0046', msg) from None

        if namespace == XMLNS_NAMESPACE:
            msg = f"cannot use the URI {XMLNS_NAMESPACE!r}!r in an EQName"
            raise self.error('XQST0070', msg)

        self.parser.advance()
        if not self.parser.next_token.label.endswith('function'):
            self.parser.expected_next('(name)', '*')
        self.parser.next_token.bind_namespace(namespace)

        cls: type[XPathToken] = self.parser.symbol_table['(string)']
        self[:] = cls(self.parser, namespace), self.parser.expression(90)

        if self[0].value:
            self.name = self[1].name = self.value = f'{{{self[0].value}}}{self[1].value}'
        elif self[1].value == '*':
            self.name = self[1].name = self.value = '{}*'
        else:
            self.name = self[1].name = self.value = cast(str, self[1].value)
        return self

    def evaluate(self, context: ta.ContextType = None) -> ta.ValueType:
        if self[1].label.endswith('function'):
            return self[1].evaluate(context)
        return xlist(self.select(context))

    def select(self, context: ta.ContextType = None) -> Iterator[ta.ItemType]:
        if self[1].label.endswith('function'):
            result = self[1].evaluate(context)
            if isinstance(result, list):
                yield from result
            else:
                yield result
            return
        elif context is None:
            raise self.missing_context()

        if isinstance(self.value, str):
            yield from context.iter_matching_nodes(self.name)


class VariableToken(XPathToken):
    """A token for parsing a variable reference."""

    symbol = lookup_name = '$'
    label = 'variable'
    bp = lbp = rbp = 90
    value: str

    def __str__(self) -> str:
        if not self._items:
            return 'unparsed variable reference'
        return f'${self[0].value} variable'

    def nud(self: XPathToken) -> XPathToken:
        self.parser.expected_next('(name)', 'Q{')
        self[:] = self.parser.expression(rbp=90),
        self.value = self[0].value
        return self

    def evaluate(self: XPathToken, context: ta.ContextType = None) -> ta.ValueType:
        if context is None:
            raise self.missing_context()

        assert isinstance(self.value, str)
        varname: str = self.value
        if varname.startswith('{'):
            expanded_name = varname
        else:
            try:
                expanded_name = get_expanded_name(self.value, self.parser.namespaces)
            except KeyError as err:
                raise self.error('XPST0081', "namespace prefix {} not found".format(err))

        try:
            value = context.variables[varname]
        except KeyError:
            if expanded_name in context.variables:
                value = context.variables[expanded_name]
            elif isinstance(context, XPathSchemaContext):
                try:
                    st = cast(dict[str, str], self.parser.variable_types)[varname]
                except KeyError:
                    return []

                if st[-1] in ('*', '+', '?'):
                    st, occurs = st[:-1], st[-1]
                else:
                    occurs = ''

                if self.parser.schema is not None and st.startswith('xs:'):
                    xsd_type = self.parser.schema.get_type(
                        st.replace('xs:', f'{{{XSD_NAMESPACE}}}', 1)
                    )
                    result = [x for x in get_atomic_sequence(xsd_type)]
                else:
                    result = [UntypedAtomic('1')]

                if occurs in ('', '?') and len(result) == 1:
                    return result[0]
                else:
                    return xlist(result)
            else:
                raise self.error('XPST0008', 'unknown variable %r' % str(varname))

        if isinstance(value, (tuple, list)):
            return xlist(value)
        elif value is None:
            return []
        else:
            return value


class AsteriskToken(XPathToken):

    symbol = lookup_name = '*'
    label = 'wildcard symbol'
    bp = lbp = rbp = 45

    def __str__(self) -> str:
        return f'{self.symbol!r} {self.label}'

    def nud(self) -> 'AsteriskToken':
        return self

    def led(self, left: XPathToken) -> 'AsteriskToken':
        self.label = 'operator'
        self[:] = left, self.parser.expression(rbp=45)
        return self

    def evaluate(self, context: ta.ContextType = None) -> ta.ValueType:
        op1: ta.ArithmeticType | None
        op2: ta.ArithmeticType
        if self:
            op1, op2 = self.get_operands(context, cls=ArithmeticProxy)
            if op1 is None:
                return []
            try:
                if isinstance(op2, (YearMonthDuration, DayTimeDuration)):
                    return op2 * op1
                return op1 * op2  # type:ignore[operator]
            except TypeError as err:
                if isinstance(context, XPathSchemaContext):
                    return []

                if isinstance(op1, (float, Decimal)):
                    if math.isnan(op1):
                        raise self.error('FOCA0005') from None
                    elif math.isinf(op1):
                        raise self.error('FODT0002') from None

                if isinstance(op2, (float, Decimal)):
                    if math.isnan(op2):
                        raise self.error('FOCA0005') from None
                    elif math.isinf(op2):
                        raise self.error('FODT0002') from None

                raise self.error('XPTY0004', err) from None
            except ValueError as err:
                if isinstance(context, XPathSchemaContext):
                    return []
                raise self.error('FOCA0005', err) from None
            except OverflowError as err:
                if isinstance(context, XPathSchemaContext):
                    return []
                elif isinstance(op1, AbstractDateTime):
                    raise self.error('FODT0001', err) from None
                elif isinstance(op1, Duration):
                    raise self.error('FODT0002', err) from None
                else:
                    raise self.error('FOAR0002', err) from None
        else:
            # This is not a multiplication operator but a wildcard select statement
            return xlist(self.select(context))

    def select(self, context: ta.ContextType = None) -> Iterator[ta.ItemType]:
        if self:
            # Product operator
            item = self.evaluate(context)
            if not isinstance(item, list):
                if context is not None:
                    context.item = item
                yield item
            elif context is not None:
                for context.item in item:
                    yield context.item
            else:
                yield from item
            return
        elif context is None:
            raise self.missing_context()

        # Wildcard literal
        if self.parser.schema is None:
            for item in context.iter_children_or_self():
                if item is None:
                    pass  # '*' wildcard doesn't match document nodes
                elif context.axis == 'attribute':
                    if isinstance(item, AttributeNode):
                        yield item
                elif isinstance(item, ElementNode):
                    yield item
        else:
            # XSD typed selection
            for item in context.iter_children_or_self():
                if context.is_principal_node_kind():
                    if isinstance(item, (ElementNode, AttributeNode)):
                        yield item


class ParentShortcutToken(XPathToken):
    symbol = lookup_name = '..'
    bp = lbp = rbp = 0
    label = 'shortcut expression'

    def nud(self) -> 'ParentShortcutToken':
        return self

    def evaluate(self, context: ta.ContextType = None) \
            -> ta.ParentNodeType | list[NoReturn]:
        if context is None:
            raise self.missing_context()

        for value in context.iter_parent():
            return value
        else:
            return []

    def select(self, context: ta.ContextType = None) -> Iterator[ta.ParentNodeType]:
        if context is None:
            raise self.missing_context()
        yield from context.iter_parent()


class ContextItemToken(XPathToken):
    """A token for parsing a context item reference."""

    symbol = lookup_name = '.'
    bp = lbp = rbp = 0
    label = 'context item expression'

    def __str__(self) -> str:
        return self.label

    def nud(self) -> 'ContextItemToken':
        return self

    def evaluate(self, context: ta.ContextType = None) -> ta.ItemType:
        if context is None:
            raise self.missing_context()
        return context.item

    def select(self, context: ta.ContextType = None) -> Iterator[ta.ItemType]:
        if context is None:
            raise self.missing_context()
        yield from context.iter_self()



#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
from typing import Any

from elementpath.helpers import collapse_white_spaces, Patterns, LazyPattern
from .any_types import AnyAtomicType

__all__ = ['NormalizedString', 'XsdToken', 'Name', 'NCName',
           'NMToken', 'Id', 'Idref', 'Language', 'Entity']


class NormalizedString(str, AnyAtomicType):
    name = 'normalizedString'
    pattern = LazyPattern('^[^\t\r]*$')

    def __new__(cls, obj: Any) -> 'NormalizedString':
        try:
            return super().__new__(cls, Patterns.normalize.sub(' ', obj))
        except TypeError:
            return super().__new__(cls, obj)

    def __init__(self, obj: Any) -> None:
        """
        :param obj: a <class 'str'> instance.
        """
        str.__init__(self)


class XsdToken(NormalizedString):
    name = 'token'
    pattern = LazyPattern(r'^[\S\xa0]*(?: [\S\xa0]+)*$')

    def __new__(cls, value: Any) -> 'XsdToken':
        if not isinstance(value, str):
            value = str(value)
        else:
            value = collapse_white_spaces(value)

        match = cls.pattern.match(value)
        if match is None:
            raise ValueError('invalid value {!r} for xs:{}'.format(value, cls.name))
        return super(NormalizedString, cls).__new__(cls, value)  # noqa


class Language(XsdToken):
    name = 'language'
    pattern = LazyPattern(r'^[a-zA-Z]{1,8}(-[a-zA-Z0-9]{1,8})*$')

    def __new__(cls, value: Any) -> 'Language':
        if isinstance(value, bool):
            value = 'true' if value else 'false'
        elif not isinstance(value, str):
            value = str(value)
        else:
            value = collapse_white_spaces(value)

        match = cls.pattern.match(value)
        if match is None:
            raise ValueError('invalid value {!r} for xs:{}'.format(value, cls.name))
        return super(NormalizedString, cls).__new__(cls, value)  # noqa


class Name(XsdToken):
    name = 'Name'
    pattern = LazyPattern(r'^(?:[^\d\W]|:)[\w.\-:\u00B7\u0300-\u036F\u203F\u2040]*$')


class NCName(Name):
    name = 'NCName'
    pattern = LazyPattern(r'^[^\d\W][\w.\-\u00B7\u0300-\u036F\u203F\u2040]*$')


class Id(NCName):
    name = 'ID'


class Idref(NCName):
    name = 'IDREF'


class Entity(NCName):
    name = 'ENTITY'


class NMToken(XsdToken):
    name = 'NMTOKEN'
    pattern = LazyPattern(r'^[\w.\-:\u00B7\u0300-\u036F\u203F\u2040]+$')



#
# Copyright (c), 2025-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
"""Fallback module for generating Unicode categories if version doesn't match."""
from sys import maxunicode
from typing import Union
from unicodedata import category

from .unicode_subsets import UnicodeSubset

UNICODE_CATEGORIES = (
    'C', 'Cc', 'Cf', 'Cs', 'Co', 'Cn',
    'L', 'Lu', 'Ll', 'Lt', 'Lm', 'Lo',
    'M', 'Mn', 'Mc', 'Me',
    'N', 'Nd', 'Nl', 'No',
    'P', 'Pc', 'Pd', 'Ps', 'Pe', 'Pi', 'Pf', 'Po',
    'S', 'Sm', 'Sc', 'Sk', 'So',
    'Z', 'Zs', 'Zl', 'Zp'
)


def get_unicodedata_categories() -> dict[str, UnicodeSubset]:
    """
    Extracts Unicode categories information from unicodedata library or from normative
    raw data. Each category is represented with an ordered list containing code points
    and code point ranges.

    :return: a dictionary with category names as keys and lists as values.
    """
    categories: dict[str, list[Union[int, tuple[int, int]]]]
    categories = {k: [] for k in UNICODE_CATEGORIES}

    major_category = 'C'
    major_start_cp, major_next_cp = 0, 1

    minor_category = 'Cc'
    minor_start_cp, minor_next_cp = 0, 1

    for cp in range(maxunicode + 1):
        cat = category(chr(cp))
        if cat[0] != major_category:
            if cp > major_next_cp:
                categories[major_category].append((major_start_cp, cp))
            else:
                categories[major_category].append(major_start_cp)

            major_category = cat[0]
            major_start_cp, major_next_cp = cp, cp + 1

        if cat != minor_category:
            if cp > minor_next_cp:
                categories[minor_category].append((minor_start_cp, cp))
            else:
                categories[minor_category].append(minor_start_cp)

            minor_category = cat
            minor_start_cp, minor_next_cp = cp, cp + 1

    else:
        if major_next_cp == maxunicode + 1:
            categories[major_category].append(major_start_cp)
        else:
            categories[major_category].append((major_start_cp, maxunicode + 1))

        if minor_next_cp == maxunicode + 1:
            categories[minor_category].append(minor_start_cp)
        else:
            categories[minor_category].append((minor_start_cp, maxunicode + 1))

    return {k: UnicodeSubset(v) for k, v in categories.items()}



#
# Copyright (c), 2016-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
"""
This module defines a class for handling Unicode subsets with less usage of memory.
"""
import sys
import warnings
from collections import defaultdict
from collections.abc import Callable, Iterable, Iterator, MutableSet
from functools import wraps
from sys import maxunicode
from types import ModuleType
from typing import cast, Optional, Union
from unicodedata import unidata_version

from .codepoints import RegexError, CodePoint, code_point_order, \
    code_point_repr, iter_code_points, iterparse_character_subset
from . import unicode_blocks
from . import unicode_categories

__all__ = ['UnicodeSubset', 'UnicodeData', 'install_unicode_data', 'unicode_version',
           'lazy_subset', 'unicode_subset', 'unicode_category', 'unicode_block']

UNICODE_VERSIONS = (
    '16.0.0', '15.1.0', '15.0.0', '14.0.0', '13.0.0', '12.1.0', '12.0.0', '11.0.0',
    '10.0.0', '9.0.0', '8.0.0', '7.0.0', '6.3.0', '6.2.0', '6.1.0', '6.0.0',
    '5.2.0', '5.1.0', '5.0.0', '4.1.0', '4.0.1', '4.0.0', '3.2.0', '3.1.1',
    '3.1.0', '3.0.1', '3.0.0', '2.1.9', '2.1.8', '2.1.5', '2.1.2', '2.0.0'
)


CodePointsArgType = Union[None, str, 'UnicodeSubset', list[CodePoint], Iterable[CodePoint]]


class UnicodeSubset(MutableSet[CodePoint]):
    """
    Represents a subset of Unicode code points, implemented with an ordered list of
    integer values and ranges. Codepoints can be added or discarded using sequences
    of integer values and ranges or with strings equivalent to regex character set.

    :param codepoints: a sequence of integer values and ranges, another UnicodeSubset \
    instance ora a string equivalent of a regex character set.
    """
    __slots__ = '_codepoints',
    _codepoints: list[CodePoint]

    def __init__(self, codepoints: CodePointsArgType = None) -> None:
        if not codepoints:
            self._codepoints = list()
        elif isinstance(codepoints, list):
            self._codepoints = sorted(codepoints, key=code_point_order)
        elif isinstance(codepoints, UnicodeSubset):
            self._codepoints = codepoints._codepoints.copy()
        else:
            self._codepoints = list()
            self.update(codepoints)

    @property
    def codepoints(self) -> list[CodePoint]:
        return self._codepoints

    @codepoints.setter
    def codepoints(self, codepoints: Iterable[CodePoint]) -> None:
        self._codepoints = sorted(codepoints, key=code_point_order)

    def __repr__(self) -> str:
        return '%s(%r)' % (self.__class__.__name__, str(self))

    def __str__(self) -> str:
        return ''.join(code_point_repr(cp) for cp in self._codepoints)

    def copy(self) -> 'UnicodeSubset':
        return self.__copy__()

    def __copy__(self) -> 'UnicodeSubset':
        subset = self.__class__()
        subset._codepoints = self._codepoints.copy()
        return subset

    def __reversed__(self) -> Iterator[int]:
        for item in reversed(self._codepoints):
            if isinstance(item, int):
                yield item
            else:
                yield from reversed(range(item[0], item[1]))

    def complement(self) -> Iterator[CodePoint]:
        last_cp = 0
        for cp in self._codepoints:
            if isinstance(cp, int):
                cp0 = cp
                cp1 = cp + 1
            else:
                cp0, cp1 = cp

            diff = cp0 - last_cp
            if diff > 2:
                yield last_cp, cp0
            elif diff == 2:
                yield last_cp
                yield last_cp + 1
            elif diff == 1:
                yield last_cp
            elif diff:
                raise ValueError("unordered code points found in {!r}".format(self))
            last_cp = cp1

        if last_cp < maxunicode:
            yield last_cp, maxunicode + 1
        elif last_cp == maxunicode:
            yield maxunicode

    def iter_characters(self) -> Iterator[str]:
        return map(chr, self.__iter__())

    #
    # MutableSet's abstract methods implementation
    def __contains__(self, value: object) -> bool:
        if not isinstance(value, int):
            try:
                value = ord(value)  # type: ignore[arg-type]
            except TypeError:
                return False

        for cp in self._codepoints:
            if not isinstance(cp, int):
                if cp[0] > value:
                    return False
                elif cp[1] <= value:
                    continue
                else:
                    return True
            elif cp > value:
                return False
            elif cp == value:
                return True
        return False

    def __iter__(self) -> Iterator[int]:
        for cp in self._codepoints:
            if isinstance(cp, int):
                yield cp
            else:
                yield from range(*cp)

    def __len__(self) -> int:
        k = 0
        for _ in self:
            k += 1
        return k

    def update(self, *others: Union[str, Iterable[CodePoint]]) -> None:
        for value in others:
            if isinstance(value, str):
                for cp in iter_code_points(iterparse_character_subset(value), reverse=True):
                    self.add(cp)
            else:
                for cp in iter_code_points(value, reverse=True):
                    self.add(cp)

    def add(self, value: CodePoint) -> None:
        if isinstance(value, int):
            if 0 <= value <= maxunicode:
                start_cp = value
                end_cp = value + 1
            else:
                raise ValueError(f"{value!r} is not a Unicode code point value")

        elif 0 <= value[0] < value[1] <= maxunicode + 1:
            start_cp, end_cp = value
        else:
            raise ValueError(f"{value!r} is not a Unicode code point range")

        code_points = self._codepoints
        last_index = len(code_points) - 1
        for k, cp in enumerate(code_points):
            if isinstance(cp, int):
                cp0 = cp
                cp1 = cp + 1
            else:
                cp0, cp1 = cp

            if end_cp < cp0:
                code_points.insert(k, value)
            elif start_cp > cp1:
                continue
            elif end_cp > cp1:
                if k == last_index:
                    code_points[k] = min(cp0, start_cp), end_cp
                else:
                    next_cp = code_points[k + 1]
                    higher_bound = next_cp if isinstance(next_cp, int) else next_cp[0]
                    if end_cp <= higher_bound:
                        code_points[k] = min(cp0, start_cp), end_cp
                    else:
                        code_points[k] = min(cp0, start_cp), higher_bound
                        start_cp = higher_bound
                        continue
            elif start_cp < cp0:
                code_points[k] = start_cp, cp1
            break
        else:
            self._codepoints.append(value)

    def difference(self, other: 'UnicodeSubset') -> 'UnicodeSubset':
        subset = self.__copy__()
        subset.difference_update(other)
        return subset

    def difference_update(self, *others: Union[str, Iterable[CodePoint]]) -> None:
        for value in others:
            if isinstance(value, str):
                for cp in iter_code_points(iterparse_character_subset(value), reverse=True):
                    self.discard(cp)
            else:
                for cp in iter_code_points(value, reverse=True):
                    self.discard(cp)

    def discard(self, value: CodePoint) -> None:
        if isinstance(value, int):
            if 0 <= value <= maxunicode:
                start_cp = value
                end_cp = value + 1
            else:
                raise ValueError(f"{value!r} is not a Unicode code point value")

        elif 0 <= value[0] < value[1] <= maxunicode + 1:
            start_cp, end_cp = value
        else:
            raise ValueError(f"{value!r} is not a Unicode code point range")

        codepoints = self._codepoints
        for k in reversed(range(len(codepoints))):
            cp = codepoints[k]
            if isinstance(cp, int):
                cp0 = cp
                cp1 = cp + 1
            else:
                cp0, cp1 = cp

            if start_cp >= cp1:
                break
            elif end_cp >= cp1:
                if start_cp <= cp0:
                    del codepoints[k]
                elif start_cp - cp0 > 1:
                    codepoints[k] = cp0, start_cp
                else:
                    codepoints[k] = cp0
            elif end_cp > cp0:
                if start_cp <= cp0:
                    if cp1 - end_cp > 1:
                        codepoints[k] = end_cp, cp1
                    else:
                        codepoints[k] = cp1 - 1
                else:
                    if cp1 - end_cp > 1:
                        codepoints.insert(k + 1, (end_cp, cp1))
                    else:
                        codepoints.insert(k + 1, cp1 - 1)
                    if start_cp - cp0 > 1:
                        codepoints[k] = cp0, start_cp
                    else:
                        codepoints[k] = cp0

    #
    # MutableSet's mixin methods override
    def clear(self) -> None:
        del self._codepoints[:]

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Iterable):
            return NotImplemented
        elif isinstance(other, UnicodeSubset):
            return self._codepoints == other._codepoints
        else:
            return self._codepoints == other

    def __ior__(self, other: object) -> 'UnicodeSubset':
        if not isinstance(other, Iterable):
            return NotImplemented
        elif isinstance(other, UnicodeSubset):
            other = reversed(other._codepoints)
        elif isinstance(other, str):
            other = reversed(UnicodeSubset(other)._codepoints)
        else:
            other = iter_code_points(other, reverse=True)

        for cp in other:
            self.add(cp)
        return self

    def __or__(self, other: object) -> 'UnicodeSubset':
        obj = self.__copy__()
        return obj.__ior__(other)

    def __isub__(self, other: object) -> 'UnicodeSubset':
        if not isinstance(other, Iterable):
            return NotImplemented
        elif isinstance(other, UnicodeSubset):
            other = reversed(other._codepoints)
        elif isinstance(other, str):
            other = reversed(UnicodeSubset(other)._codepoints)
        else:
            other = iter_code_points(other, reverse=True)

        for cp in other:
            self.discard(cp)
        return self

    def __sub__(self, other: object) -> 'UnicodeSubset':
        obj = self.__copy__()
        return obj.__isub__(other)

    __rsub__ = __sub__

    def __iand__(self, other: object) -> 'UnicodeSubset':
        if not isinstance(other, Iterable):
            return NotImplemented

        for value in (self - other):
            self.discard(value)
        return self

    def __and__(self, other: object) -> 'UnicodeSubset':
        obj = self.__copy__()
        return obj.__iand__(other)

    def __ixor__(self, other: object) -> 'UnicodeSubset':
        if other is self:
            self.clear()
            return self
        elif not isinstance(other, Iterable):
            return NotImplemented
        elif not isinstance(other, UnicodeSubset):
            other = UnicodeSubset(cast(Union[str, Iterable[CodePoint]], other))

        for value in other:
            if value in self:
                self.discard(value)
            else:
                self.add(value)
        return self

    def __xor__(self, other: object) -> 'UnicodeSubset':
        obj = self.__copy__()
        return obj.__ixor__(other)


def iterparse_unicode_data(url: str) -> Iterator[tuple[int, str]]:
    """Iterate UnicodeData.txt source giving back codepoints and categories."""
    from urllib.request import urlopen

    with urlopen(url) as res:
        prev_cp = -1

        for line in res.readlines():
            fields = line.split(b';')
            cp = int(fields[0], 16)
            cat = fields[2].decode('utf-8')

            if cp - prev_cp > 1:
                if fields[1].endswith(b', Last>'):
                    # Ranges of codepoints expressed with First and then Last
                    for x in range(prev_cp + 1, cp):
                        yield x, cat
                else:
                    # For default is 'Cn' that means 'Other, not assigned'
                    for x in range(prev_cp + 1, cp):
                        yield x, 'Cn'

            prev_cp = cp
            yield cp, cat

    while cp < maxunicode:
        cp += 1
        yield cp, 'Cn'


def get_categories_from_url(url: str) -> dict[str, UnicodeSubset]:
    categories: dict[str, list[CodePoint]] = defaultdict(list)

    major_category = 'C'
    major_start_cp, major_next_cp = 0, 1

    minor_category = 'Cc'
    minor_start_cp, minor_next_cp = 0, 1

    for cp, cat in iterparse_unicode_data(url):

        if cat[0] != major_category:
            if cp > major_next_cp:
                categories[major_category].append((major_start_cp, cp))
            else:
                categories[major_category].append(major_start_cp)

            major_category = cat[0]
            major_start_cp, major_next_cp = cp, cp + 1

        if cat != minor_category:
            if cp > minor_next_cp:
                categories[minor_category].append((minor_start_cp, cp))
            else:
                categories[minor_category].append(minor_start_cp)

            minor_category = cat
            minor_start_cp, minor_next_cp = cp, cp + 1

    else:
        if major_next_cp == maxunicode + 1:
            categories[major_category].append(major_start_cp)
        else:
            categories[major_category].append((major_start_cp, maxunicode + 1))

        if minor_next_cp == maxunicode + 1:
            categories[minor_category].append(minor_start_cp)
        else:
            categories[minor_category].append((minor_start_cp, maxunicode + 1))

    return {k: UnicodeSubset(v) for k, v in categories.items()}


def get_categories(version_info: tuple[int, ...], module: ModuleType) -> dict[str, UnicodeSubset]:
    categories = {k: v.copy() for k, v in module.UNICODE_CATEGORIES.items()}
    for name in module.__dict__:
        if not name.startswith('DIFF_CATEGORIES_VER_'):
            continue

        diff_version = name[20:].replace('_', '.')
        if version_info < tuple(int(x) for x in diff_version.split('.')):
            break

        for k, (exclude_cps, insert_cps) in getattr(unicode_categories, name).items():
            values = []
            additional = iter(insert_cps)
            cpa = next(additional, None)
            cpa_int = cpa[0] if isinstance(cpa, tuple) else cpa

            for cp in categories[k]:
                if cp in exclude_cps:
                    continue

                cp_int = cp[0] if isinstance(cp, tuple) else cp
                while cpa_int is not None and cpa_int <= cp_int:
                    values.append(cpa)
                    cpa = next(additional, None)
                    cpa_int = cpa[0] if isinstance(cpa, tuple) else cpa
                else:
                    values.append(cp)
            else:
                if cpa is not None:
                    values.append(cpa)
                    values.extend(additional)

            categories[k] = values

    return {k: UnicodeSubset(v) for k, v in categories.items()}


class UnicodeData:

    _blocks: dict[str, Union[str, UnicodeSubset]]

    @staticmethod
    def _unicode_block_key(name: str) -> str:
        return name.upper().replace(' ', '').replace('_', '').replace('-', '')

    def __init__(self, version: Optional[str] = None,
                 categories: Optional[dict[str, UnicodeSubset]] = None) -> None:
        if version is None:
            version = unidata_version
        elif version not in UNICODE_VERSIONS:
            raise ValueError("argument is not a valid Unicode version")

        self.version = version
        version_info = tuple(int(x) for x in version.split('.'))
        if categories is not None:
            self._categories = categories
        elif self.version in unicode_categories.UNICODE_VERSIONS:
            self._categories = get_categories(version_info, unicode_categories)
        else:
            from . import categories_fallback

            py_version = '.'.join(str(n) for n in sys.version_info[:2])
            msg = (f"unexpected Unicode version {version} for Python {py_version}, "
                   f"use unicodedata module for building categories map.")
            warnings.warn(UnicodeWarning(msg))

            self._categories = categories_fallback.get_unicodedata_categories()

        # Build blocks dict for version
        superseded_blocks = []
        blocks = unicode_blocks.UNICODE_BLOCKS_VER_2_0_0.copy()

        for name in unicode_blocks.__dict__:  # noqa
            if name.startswith('UPDATE_BLOCKS_VER_'):
                diff_version = name[18:].replace('_', '.')
                if version_info < tuple(int(x) for x in diff_version.split('.')):
                    break
                blocks.update(getattr(unicode_blocks, name))

            elif name.startswith('REMOVED_BLOCKS_VER_'):
                diff_version = name[19:].replace('_', '.')
                if version_info < tuple(int(x) for x in diff_version.split('.')):
                    break
                superseded_blocks.extend(getattr(unicode_blocks, name))

        # Following naming rules: https://www.w3.org/TR/xmlschema11-2/#cces-blockesc
        self._blocks = {k.replace(' ', '').replace('_', ''): v for k, v in blocks.items()}

        # Additional map for lookup using normalization for Unicode naming rules,
        # doesn't include superseded blocks.
        self._unicode_blocks = {
            k.upper().replace(' ', '').replace('_', '').replace('-', ''): k
            for k in blocks if k not in superseded_blocks
        }

    def category(self, name: str) -> UnicodeSubset:
        return self._categories[name]

    def block(self, name: str, normalize: bool = False) -> UnicodeSubset:
        if normalize:
            key = name.upper().replace(' ', '').replace('_', '').replace('-', '')
            try:
                name = self._unicode_blocks[key]
            except KeyError:
                if key != 'NOBLOCK':
                    raise
                name = key

        try:
            subset = self._blocks[name]
        except KeyError:
            if name != 'NoBlock':
                raise

            # Define the special block "No_Block", that contains all the other codepoints not
            # belonging to a defined block (https://www.unicode.org/Public/UNIDATA/Blocks.txt)
            no_block = UnicodeSubset([(0, maxunicode + 1)])
            for v in self._blocks.values():
                no_block -= v
            self._blocks['NoBlock'] = no_block
            self._unicode_blocks['NOBLOCK'] = 'NoBlock'
            return no_block

        else:
            if not isinstance(subset, UnicodeSubset):
                subset = self._blocks[name] = UnicodeSubset(subset)
            return subset


###
# Installed Unicode Data instance and accessors
__unicode_data = UnicodeData()

# Simple cache for Unicode subsets defined using callables with no-arguments, that
# can include subsets defined on versioned Unicode data or fixed codepoints. This
# cache is cleared if Unicode data is reinstalled.
__subsets_cache: dict[Callable[[], UnicodeSubset], UnicodeSubset] = {}


def install_unicode_data(version: Optional[str] = None,
                         name_or_url: Optional[str] = None) -> None:
    """
    Install a different version of UnicodeData. For default the package installs the version
    that matches `unicodedata.unidata_version`. Call without parameters to restore the
    default version.

    :param version: Unicode version to install. It's required if a *name_or_url* is provided.
    :param name_or_url: Import name of an additional module or a URL to raw UnicodeData.txt.
    """
    global __unicode_data

    if name_or_url is None:
        __unicode_data = UnicodeData(version)
    elif version is None:
        raise TypeError("you must specify a version to install")
    elif name_or_url.endswith('unicode_categories'):
        import importlib
        module = importlib.import_module(name_or_url)
        version_info = tuple(int(x) for x in version.split('.'))
        categories = get_categories(version_info, module)
        __unicode_data = UnicodeData(version, categories)
    else:
        categories = get_categories_from_url(name_or_url)
        __unicode_data = UnicodeData(version, categories)

    __subsets_cache.clear()


def unicode_version() -> str:
    """Returns the installed UnicodeData version."""
    return __unicode_data.version


def lazy_subset(func: Callable[[], UnicodeSubset]) -> Callable[[], UnicodeSubset]:
    """
    Defines a lazy UnicodeSubset wrapping its definition in a callable with no arguments.
    """
    @wraps(func)
    def wrapper() -> UnicodeSubset:
        try:
            return __subsets_cache[func]
        except KeyError:
            __subsets_cache[func] = func()
            return __subsets_cache[func]

    return wrapper


def unicode_subset(name: str) -> UnicodeSubset:
    """Retrieve a Unicode subset by name, raising a RegexError if it cannot be retrieved."""
    if name[:2] == 'Is':
        try:
            return __unicode_data.block(name[2:])
        except KeyError:
            raise RegexError(f"{name!r} doesn't match any Unicode block")
    else:
        try:
            return __unicode_data.category(name)
        except KeyError:
            raise RegexError(f"{name!r} doesn't match any Unicode category")


def unicode_category(name: str) -> UnicodeSubset:
    """
    Returns the Unicode Character Category subset addressed by the provided name, raising a
    KeyError if it's not found.
    """
    return __unicode_data.category(name)


def unicode_block(name: str, normalize: bool = False) -> UnicodeSubset:
    """
    Returns the Unicode block subset addressed by the provided name, raising a KeyError
    if it's not found. For default the lookup is done following the XSD naming rules for
    blocks and keeping superseded blocks (e.g. Greek), otherwise the name is normalized
    following the Unicode standard rules, without considering the casing, spaces, hyphens
    and underscores and the lookup is restricted to blocks defined on installed version.
    """
    return __unicode_data.block(name, normalize)



#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
from collections import deque
from collections.abc import Iterator
from typing import cast, Any, Optional, Union

import elementpath.aliases as ta
from elementpath.exceptions import ElementPathTypeError
from elementpath.protocols import LxmlElementProtocol, DocumentProtocol, \
    LxmlDocumentProtocol, XsdElementProtocol, DocumentType, ElementType, \
    SchemaElemType
from elementpath.etree import is_etree_document, is_etree_element, is_etree_element_instance
from elementpath.xpath_nodes import TextNode, ElementNode, SchemaElementNode, DocumentNode, \
    EtreeElementNode, EtreeDocumentNode, CommentNode, ProcessingInstructionNode

__all__ = ['get_node_tree', 'build_node_tree', 'build_lxml_node_tree', 'build_schema_node_tree']

ElementTreeRootType = Union[DocumentType, ElementType]
LxmlRootType = Union[LxmlDocumentProtocol, LxmlElementProtocol]


def is_schema(obj: Any) -> bool:
    return hasattr(obj, 'xsd_version') and hasattr(obj, 'maps') and not hasattr(obj, 'parent')


def get_node_tree(root: ta.RootArgType,
                  namespaces: Optional[ta.NamespacesType] = None,
                  uri: Optional[str] = None,
                  fragment: Optional[bool] = None) -> ta.RootNodeType:
    """
    Returns a tree of XPath nodes that wrap the provided root tree.

    :param root: an Element or an ElementTree or a schema or a schema element.
    :param namespaces: an optional mapping from prefixes to namespace URIs, \
    Ignored if root is a lxml etree or a schema structure.
    :param uri: an optional URI associated with the root element or the document.
    :param fragment: if `True` is provided the root is considered a fragment. In this \
    case if `root` is an ElementTree instance skips it and use the root Element. If \
    `False` is provided creates a dummy document when the root is an Element instance. \
    For default the root node kind is preserved.
    """
    if isinstance(root, (DocumentNode, ElementNode)):
        if uri is not None and root.uri is None:
            root.tree.uri = uri

        if fragment:
            if isinstance(root, DocumentNode):
                return root.getroot()
        elif fragment is False and \
                isinstance(root, ElementNode) and \
                is_etree_element_instance(root.value):
            return root.get_document_node()

        return root

    if not is_etree_document(root) and \
            (not is_etree_element(root) or callable(cast(ElementType, root).tag)):
        msg = "invalid root {!r}, an Element or an ElementTree or a schema node required"
        raise ElementPathTypeError(msg.format(root))
    elif hasattr(root, 'xpath'):
        # a lxml element tree data
        return build_lxml_node_tree(
            cast(LxmlRootType, root), uri, fragment
        )
    elif hasattr(root, 'xsd_version') and hasattr(root, 'maps'):
        # a schema or a schema node
        return build_schema_node_tree(
            cast(SchemaElemType, root), uri
        )
    else:
        return build_node_tree(root, namespaces, uri, fragment)


def build_node_tree(root: ElementTreeRootType,
                    namespaces: Optional[ta.NamespacesType] = None,
                    uri: Optional[str] = None,
                    fragment: Optional[bool] = None) -> ta.RootNodeType:
    """
    Returns a tree of XPath nodes that wrap the provided root tree.

    :param root: an Element or an ElementTree.
    :param namespaces: an optional mapping from prefixes to namespace URIs.
    :param uri: an optional URI associated with the document or the root element.
    :param fragment: if `True` is provided the root is considered a fragment. In this \
    case if `root` is an ElementTree instance skips it and use the root Element. If \
    `False` is provided creates a dummy document when the root is an Element instance. \
    For default the root node kind is preserved.
    """
    elem: ElementType
    parent: Any
    child: ta.ChildNodeType
    children: Iterator[Any]
    document: Optional[DocumentProtocol]

    position = 1
    if namespaces is None:
        namespaces = {}
    ns_pos_offset = len(namespaces) + int('xml' not in namespaces) + 1

    if hasattr(root, 'parse'):
        document = cast(DocumentProtocol, root)
        root_elem = document.getroot()
    else:
        document = None
        root_elem = root

    if fragment and root_elem is not None:
        document = None  # Explicitly requested a fragment: don't create a document node

    if document is not None:
        document_node = EtreeDocumentNode(document, uri, position)
        position += 1

        if root_elem is None:
            return document_node

        elem = root_elem
        root_node = EtreeElementNode(elem, document_node, position)
    else:
        assert root_elem is not None
        document_node = None
        elem = root_elem

        root_node = EtreeElementNode(elem, None, position)
        if uri is not None:
            root_node.uri = uri

    root_node.tree.namespaces = namespaces
    position += ns_pos_offset + len(elem.attrib)
    if elem.text is not None:
        TextNode(elem.text, root_node, position)
        position += 1

    children = iter(elem)
    iterators: deque[Any] = deque()
    ancestors: deque[Any] = deque()
    parent = root_node

    while True:
        for elem in children:
            if not callable(elem.tag):
                child = EtreeElementNode(elem, parent, position)
                position += ns_pos_offset + len(elem.attrib)

                if elem.text is not None:
                    TextNode(elem.text, child, position)
                    position += 1

            elif elem.tag.__name__ == 'Comment':
                child = CommentNode(elem, parent, position)
                position += 1
            else:
                child = ProcessingInstructionNode(elem, None, parent, position)
                position += 1

            if len(elem):
                ancestors.append(parent)
                parent = child
                iterators.append(children)
                children = iter(elem)
                break

            if elem.tail is not None:
                TextNode(elem.tail, parent, position)
                position += 1
        else:
            try:
                children, parent = iterators.pop(), ancestors.pop()
            except IndexError:
                if document_node is not None:
                    return document_node
                elif fragment is False and \
                        isinstance(root_node, ElementNode) and \
                        is_etree_element_instance(root_node.elem):
                    return root_node.get_document_node()
                else:
                    return root_node
            else:
                if (tail := parent.children[-1].elem.tail) is not None:
                    TextNode(tail, parent, position)
                    position += 1


def build_lxml_node_tree(root: LxmlRootType,
                         uri: Optional[str] = None,
                         fragment: Optional[bool] = None) -> ta.RootNodeType:
    """
    Returns a tree of XPath nodes that wrap the provided lxml root tree.

    :param root: a lxml Element or a lxml ElementTree.
    :param uri: an optional URI associated with the document or the root element.
    :param fragment: if `True` is provided the root is considered a fragment. In this \
    case if `root` is an ElementTree instance skips it and use the root Element. If \
    `False` is provided creates a dummy document when the root is an Element instance. \
    For default the root node kind is preserved.
    """
    root_node: ta.RootNodeType
    document: Optional[LxmlDocumentProtocol]
    parent: Any
    child: ta.ChildNodeType
    children: Iterator[Any]

    position = 1

    if fragment:
        document = None  # Explicitly requested a fragment: don't create a document node
    elif hasattr(root, 'parse'):
        document = cast(LxmlDocumentProtocol, root)
    elif fragment is False or root.getparent() is None and (
            any(True for _sibling in root.itersiblings(preceding=True)) or
            any(True for _sibling in root.itersiblings())):
        # Despite a document is not explicitly requested create a dummy document
        # because the root element has siblings
        document = root.getroottree()
    else:
        document = None

    if document is not None:
        document_node = EtreeDocumentNode(document, uri, position)
        position += 1

        root_elem = document.getroot()
        if root_elem is None:
            return document_node

        # Add root siblings (comments and processing instructions)
        for elem in reversed([x for x in root_elem.itersiblings(preceding=True)]):
            assert callable(elem.tag)
            if elem.tag.__name__ == 'Comment':
                CommentNode(elem, document_node, position)
                position += 1
            else:
                ProcessingInstructionNode(elem, None, document_node, position)
                position += 1

        root_node = EtreeElementNode(root_elem, document_node, position)
    else:
        if hasattr(root, 'parse'):
            root_elem = cast(LxmlDocumentProtocol, root).getroot()
        else:
            root_elem = root

        if root_elem is None:
            if fragment:
                msg = "requested a fragment of an empty ElementTree document"
            else:
                msg = "root argument is neither an lxml ElementTree nor a lxml Element"
            raise ElementPathTypeError(msg)

        document_node = None
        root_node = EtreeElementNode(root_elem)
        if uri is not None:
            root_node.uri = uri

    # Complete the root element node build
    if 'xml' in root_elem.nsmap:
        position += len(root_elem.nsmap) + len(root_elem.attrib) + 1
    else:
        position += len(root_elem.nsmap) + len(root_elem.attrib) + 2

    if root_elem.text is not None:
        TextNode(root_elem.text, root_node, position)
        position += 1

    children = iter(root_elem)
    iterators: deque[Any] = deque()
    ancestors: deque[Any] = deque()
    parent = root_node

    while True:
        for elem in children:
            if not callable(elem.tag):
                child = EtreeElementNode(elem, parent, position)
                if 'xml' in elem.nsmap:
                    position += len(elem.nsmap) + len(elem.attrib) + 1
                else:
                    position += len(elem.nsmap) + len(elem.attrib) + 2

                if elem.text is not None:
                    TextNode(elem.text, child, position)
                    position += 1

            elif elem.tag.__name__ == 'Comment':
                child = CommentNode(elem, parent, position)
                position += 1
            else:
                child = ProcessingInstructionNode(elem, None, parent, position)
                position += 1

            if len(elem):
                ancestors.append(parent)
                parent = child
                iterators.append(children)
                children = iter(elem)
                break

            if elem.tail is not None:
                TextNode(elem.tail, parent, position)
                position += 1
        else:
            try:
                children, parent = iterators.pop(), ancestors.pop()
            except IndexError:
                if document_node is None:
                    return root_node

                # Add root following siblings (comments and processing instructions)
                for elem in root_elem.itersiblings():
                    assert callable(elem.tag)
                    if elem.tag.__name__ == 'Comment':
                        CommentNode(elem, document_node, position)
                        position += 1
                    else:
                        ProcessingInstructionNode(elem, None, document_node, position)
                        position += 1

                return document_node
            else:
                if (tail := parent.children[-1].elem.tail) is not None:
                    TextNode(tail, parent, position)
                    position += 1


def build_schema_node_tree(root: SchemaElemType,
                           uri: Optional[str] = None,
                           elements: Optional[ta.ElementMapType] = None,
                           global_elements: Optional[list[ta.ChildNodeType]] = None) \
        -> SchemaElementNode:
    """
    Returns a graph of XPath nodes that wrap the provided XSD schema structure.
    The elements dictionary is shared between all nodes to keep all of them,
    globals and local, linked in a single structure.

    :param root: a schema or a schema element.
    :param uri: an optional URI associated with the root element.
    :param elements: a shared map from XSD elements to tree nodes. Provided for \
    linking together parts of the same schema or other schemas.
    :param global_elements: a list for schema global elements, used for linking \
    the elements declared by reference.
    """
    parent: SchemaElementNode
    elem: XsdElementProtocol
    child: SchemaElementNode
    children: Iterator[Any]

    position = 1
    if hasattr(root, 'namespaces'):
        namespaces: ta.NamespacesType = root.namespaces
        ns_pos_offset = len(namespaces) + int('xml' not in namespaces) + 1
    else:
        namespaces = {}
        ns_pos_offset = 2

    root_node = SchemaElementNode(root, None, position)
    if elements is not None:
        elements.update(root_node.tree.elements)
        root_node.tree.elements = elements
    root_node.tree.namespaces = namespaces
    position += ns_pos_offset + len(root.attrib)

    if global_elements is not None:
        global_elements.append(root_node)
    elif is_schema(root):
        global_elements = root_node.children
    else:
        # Track global elements even if the initial root is not a schema to avoid circularity
        global_elements = []

    local_nodes = {root: root_node}  # Irrelevant even if it's the schema
    ref_nodes: list[SchemaElementNode] = []

    if is_schema(root):
        children = iter(e for e in root.maps.elements.values() if e.maps is root.maps)
    else:
        children = iter(root)

    iterators: list[Any] = []
    ancestors: list[Any] = []
    parent = root_node

    while True:
        for elem in children:
            child = SchemaElementNode(elem, parent, position)
            position += ns_pos_offset + len(elem.attrib)

            if elem in local_nodes:
                if elem.ref is None:
                    child.children = local_nodes[elem].children
                else:
                    ref_nodes.append(child)
            else:
                local_nodes[elem] = child
                if elem.ref is None:
                    ancestors.append(parent)
                    parent = child
                    iterators.append(children)
                    children = iter(elem)
                    break
                else:
                    ref_nodes.append(child)
        else:
            try:
                children, parent = iterators.pop(), ancestors.pop()
            except IndexError:
                # connect references to proper nodes
                for element_node in ref_nodes:
                    elem = cast(XsdElementProtocol, element_node.elem)
                    ref = cast(XsdElementProtocol, elem.ref)

                    other: Any
                    for other in global_elements:
                        if other.elem is ref:
                            element_node.ref = other
                            break
                    else:
                        # Extend node tree with other globals
                        element_node.ref = build_schema_node_tree(
                            ref, elements=root_node.elements, global_elements=global_elements
                        )

                return root_node



#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
"""
XPath 2.0 implementation - part 1 (parser class and symbols)
"""
import copy
from abc import ABCMeta
import locale
from collections.abc import Callable, MutableMapping
from urllib.parse import urlparse
from typing import Any, ClassVar, cast

import elementpath.aliases as ta

from elementpath.helpers import upper_camel_case, is_ncname
from elementpath.exceptions import ElementPathTypeError, ElementPathValueError, xpath_error
from elementpath.namespaces import XML_NAMESPACE, XSD_NAMESPACE, XPATH_FUNCTIONS_NAMESPACE, \
    XQT_ERRORS_NAMESPACE, XSD_NOTATION, XSD_ANY_ATOMIC_TYPE
from elementpath.namespaces import get_prefixed_name
from elementpath.datatypes import QName, builtin_atomic_types
from elementpath.collations import UNICODE_COLLATION_BASE_URI, UNICODE_CODEPOINT_COLLATION
from elementpath.xpath_tokens import XPathToken, ProxyToken, XPathFunction, \
    XPathConstructor, SchemaConstructor, ExternalFunction
from elementpath.sequence_types import get_function_signatures, is_sequence_type, \
    match_sequence_type
from elementpath.schema_proxy import AbstractSchemaProxy
from elementpath.xpath1 import XPath1Parser


class XPath2Parser(XPath1Parser):
    """
    XPath 2.0 expression parser class. This is the default parser used by XPath selectors.
    A parser instance represents also the XPath static context. With *variable_types* you
    can pass a dictionary with the types of the in-scope variables.
    Provide a *namespaces* dictionary argument for mapping namespace prefixes to URI inside
    expressions. If *strict* is set to `False` the parser enables also the parsing of QNames,
    like the ElementPath library. There are some additional XPath 2.0 related arguments.

    :param namespaces: a dictionary with mapping from namespace prefixes into URIs.
    :param variable_types: a dictionary with the static context's in-scope variable \
    types. It defines the associations between variables and static types.
    :param strict: if strict mode is `False` the parser enables parsing of QNames, \
    like the ElementPath library. Default is `True`.
    :param compatibility_mode: if set to `True` the parser instance works with \
    XPath 1.0 compatibility rules.
    :param default_namespace: the default namespace to apply to unprefixed names. \
    For default no namespace is applied (empty namespace '').
    :param function_namespace: the default namespace to apply to unprefixed function \
    names. For default the namespace "http://www.w3.org/2005/xpath-functions" is used.
    :param schema: the schema proxy class or instance to use for types, attributes and \
    elements lookups. If an `AbstractSchemaProxy` subclass is provided then a schema \
    proxy instance is built without the optional argument, that involves a mapping of \
    only XSD builtin types. If it's not provided the XPath 2.0 schema's related \
    expressions cannot be used.
    :param base_uri: an absolute URI maybe provided, used when necessary in the \
    resolution of relative URIs.
    :param default_collation: the default string collation to use. If not set the \
    environment's default locale setting is used.
    :param document_types: statically known documents, that is a dictionary from \
    absolute URIs onto types. Used for type check when calling the *fn:doc* function \
    with a sequence of URIs. The default type of a document is 'document-node()'.
    :param collection_types: statically known collections, that is a dictionary from \
    absolute URIs onto types. Used for type check when calling the *fn:collection* \
    function with a sequence of URIs. The default type of a collection is 'node()*'.
    :param default_collection_type: this is the type of the sequence of nodes that \
    would result from calling the *fn:collection* function with no arguments. \
    Default is 'node()*'.
    """
    version = '2.0'

    DEFAULT_NAMESPACES: ClassVar[dict[str, str]] = {
        'xml': XML_NAMESPACE,
        'xs': XSD_NAMESPACE,
        'fn': XPATH_FUNCTIONS_NAMESPACE,
        'err': XQT_ERRORS_NAMESPACE
    }

    PATH_STEP_LABELS = ('axis', 'function', 'kind test')
    PATH_STEP_SYMBOLS = {
        '(integer)', '(string)', '(float)', '(decimal)', '(name)', '*', '@', '..', '.', '(', '{'
    }

    # https://www.w3.org/TR/xpath20/#id-reserved-fn-names
    RESERVED_FUNCTION_NAMES = {
        'attribute', 'comment', 'document-node', 'element', 'empty-sequence',
        'if', 'item', 'node', 'processing-instruction', 'schema-attribute',
        'schema-element', 'text', 'typeswitch',
    }

    function_signatures: dict[tuple[QName, int], str] = XPath1Parser.function_signatures.copy()
    namespaces: dict[str, str]
    token: XPathToken
    next_token: XPathToken

    def __init__(self, namespaces: ta.NamespacesType | None = None,
                 strict: bool = True,
                 compatibility_mode: bool = False,
                 default_collation: str | None = None,
                 default_namespace: str | None = None,
                 function_namespace: str | None = None,
                 xsd_version: str | None = None,
                 schema: AbstractSchemaProxy | None = None,
                 base_uri: str | None = None,
                 variable_types: dict[str, str] | None = None,
                 document_types: dict[str, str] | None = None,
                 collection_types: ta.NamespacesType | None = None,
                 default_collection_type: str = 'node()*') -> None:

        super(XPath2Parser, self).__init__(namespaces, strict)
        self.compatibility_mode = compatibility_mode

        if default_collation is not None:
            self.default_collation = default_collation
        else:
            # Obtain the current collation locale using setlocale() with `None`.
            # Consider only configured UTF-8 encodings, otherwise keep Unicode
            # Codepoint Collation.
            _locale = locale.setlocale(locale.LC_COLLATE, None)
            if '.' in _locale:
                language_code, encoding = _locale.split('.')
                if encoding.lower() == 'utf-8':
                    self.default_collation = f'{UNICODE_COLLATION_BASE_URI}?lang={language_code}'

        self._xsd_version = xsd_version if xsd_version is not None else '1.0'

        if default_namespace is not None:
            self.default_namespace = self.namespaces[''] = default_namespace
        else:
            self.default_namespace = self.namespaces.get('', '')

        if function_namespace is not None:
            self.function_namespace = function_namespace

        if schema is None:
            pass
        elif not isinstance(schema, AbstractSchemaProxy):
            msg = "argument 'schema' must be an instance of AbstractSchemaProxy"
            raise ElementPathTypeError(msg)
        else:
            schema.bind_parser(self)

        if not variable_types:
            self.variable_types = {}
        elif all(is_sequence_type(v, self) for v in variable_types.values()):
            self.variable_types = variable_types.copy()
        else:
            raise ElementPathValueError('invalid sequence type for in-scope variable types')

        self.base_uri = None if base_uri is None else urlparse(base_uri).geturl()

        if document_types:
            if any(not is_sequence_type(v, self) for v in document_types.values()):
                raise ElementPathValueError('invalid sequence type in document_types argument')
        self.document_types = document_types

        if collection_types:
            if any(not is_sequence_type(v, self) for v in collection_types.values()):
                raise ElementPathValueError('invalid sequence type in collection_types argument')
        self.collection_types = collection_types

        if not is_sequence_type(default_collection_type, self):
            raise ElementPathValueError('invalid sequence type for '
                                        'default_collection_type argument')
        self.default_collection_type = default_collection_type

    def __str__(self) -> str:
        args = []
        if self.compatibility_mode:
            args.append('compatibility_mode=True')
        if self.default_collation != UNICODE_CODEPOINT_COLLATION:
            args.append(f'default_collation={self.default_collation!r}')
        if self.function_namespace != XPATH_FUNCTIONS_NAMESPACE:
            args.append(f'function_namespace={self.function_namespace!r}')
        if self._xsd_version != '1.0':
            args.append(f'xsd_version={self._xsd_version!r}')
        if self.schema is not None:
            args.append(f'schema={self.schema!r}')
        if self.base_uri is not None:
            args.append(f'base_uri={self.base_uri!r}')
        if self.variable_types:
            args.append(f'variable_types={self.variable_types!r}')
        if self.document_types:
            args.append(f'document_types={self.document_types!r}')
        if self.collection_types:
            args.append(f'collection_types={self.collection_types!r}')
        if self.default_collection_type != 'node()*':
            args.append(f'default_collection_type={self.default_collection_type!r}')
        if not args:
            return super().__str__()

        repr_string = super().__str__()[:-1]
        if repr_string.endswith('('):
            return f"{repr_string}{', '.join(args)})"
        return f"{repr_string}, {', '.join(args)})"

    def __getstate__(self) -> dict[str, Any]:
        state = self.__dict__.copy()
        state.pop('symbol_table', None)
        state.pop('tokenizer', None)
        return state

    @property
    def xsd_version(self) -> str:
        if self.schema is None:
            return self._xsd_version

        try:
            return self.schema.xsd_version
        except (AttributeError, NotImplementedError):
            return self._xsd_version

    def advance(self, *symbols: str,  message: str | None = None) -> XPathToken:
        super(XPath2Parser, self).advance(*symbols, message=message)

        if self.next_token.symbol == '(:':
            # Parses and consumes an XPath 2.0 comment. A comment is delimited
            # by symbols '(:' and ':)' and can be nested. The current token is
            # saved and restored after parsing the entire comment. Comments
            # cannot be inside a prefixed name ':' specification.
            self.token.unexpected(':')
            token = self.token

            comment_level = 1
            while comment_level:
                self.advance_until('(:', ':)')
                if self.next_token.symbol == ':)':
                    comment_level -= 1
                else:
                    comment_level += 1
            self.advance(':)')

            self.next_token.unexpected(':')
            self.token = token

        return self.token

    @classmethod
    def constructor(cls, symbol: str, bp: int = 90, nargs: ta.NargsType = 1,
                    sequence_types: tuple[()] | tuple[str, ...] | list[str] = (),
                    label: str | tuple[str, ...] = 'constructor function') \
            -> Callable[[Callable[..., Any]], Callable[..., Any]]:
        """
        Statically creates a constructor token class, that is registered in the globals
        of the module where the method is called.
        """
        if not sequence_types:
            assert nargs == 1
            sequence_types = ('xs:anyAtomicType?', 'xs:%s?' % symbol)

        type_name = f'{{{XSD_NAMESPACE}}}{symbol}'
        kwargs = {
            'bases': (XPathConstructor,),
            'label': label,
            'nargs': nargs,
            'lbp': bp,
            'rbp': bp,
            'sequence_types': sequence_types,
            'name': type_name,
            'type_class': builtin_atomic_types.get(type_name),
        }
        token_class = cls.register(symbol, **kwargs)

        def bind(func: Callable[..., Any]) -> Callable[..., Any]:
            method_name = func.__name__.partition('_')[0]
            if method_name != 'cast':
                raise ValueError("The function name must be 'cast' or starting with 'cast_'")
            setattr(token_class, method_name, func)
            return func
        return bind

    def _register(self, symbol: str, class_name: str, **kwargs: Any) \
            -> type[ta.XPathTokenType]:
        """
        Register/update a token class in the symbol table of a parser instance.

        :param symbol: The identifier symbol for the new class.
        :param class_name: The name of the new class.
        :param kwargs: Optional attributes/methods for the token class.
        :return: A token class.
        """
        kwargs['symbol'] = symbol
        if 'lookup_name' not in kwargs:
            lookup_name = kwargs['lookup_name'] = symbol
        else:
            lookup_name = kwargs['lookup_name']

        if 'label' not in kwargs:
            kwargs['label'] = 'symbol'

        token_class_bases = kwargs.get('bases', (self.token_base_class,))
        kwargs.update({
            '__module__': self.__module__,
            '__qualname__': class_name,
            '__return__': None
        })
        token_class = cast(
            type[ta.XPathTokenType],
            cast(object, ABCMeta(class_name, token_class_bases, kwargs))
        )

        if self.symbol_table is self.__class__.symbol_table:
            self.symbol_table = copy.copy(self.symbol_table)
        self.symbol_table[lookup_name] = token_class
        self.tokenizer = None
        return token_class

    def schema_constructor(self, atomic_type_name: str, bp: int = 90) -> type[XPathFunction]:
        """
        Dynamically register a token class for a schema atomic type constructor function.
        """
        if atomic_type_name in (XSD_ANY_ATOMIC_TYPE, XSD_NOTATION):
            raise xpath_error('XPST0080')

        symbol = get_prefixed_name(atomic_type_name, self.namespaces)
        class_name = "_%sConstructorFunction" % symbol.replace(':', '_')
        kwargs = {
            'bases': (SchemaConstructor,),
            'name': atomic_type_name,
            'nargs': 1,
            'label': 'constructor function',
            'pattern': r'\b%s(?=\s*\(|\s*\(\:.*\:\)\()' % symbol,
            'lbp': bp,
            'rbp': bp,
        }
        return cast(type[XPathFunction], self._register(symbol, class_name, **kwargs))

    def external_function(self,
                          callback: Callable[..., Any],
                          name: str | None = None,
                          prefix: str | None = None,
                          sequence_types: tuple[str, ...] = (),
                          bp: int = 90) -> type[XPathFunction]:
        """Registers a token class for an external function."""
        import inspect

        symbol = name or callback.__name__
        if not is_ncname(symbol):
            raise ElementPathValueError(f'{symbol!r} is not a name')
        elif symbol in self.RESERVED_FUNCTION_NAMES:
            raise ElementPathValueError(f'{symbol!r} is a reserved function name')

        nargs: ta.NargsType
        spec = inspect.getfullargspec(callback)
        if spec.varargs is not None:
            if spec.args:
                nargs = len(spec.args), None
            else:
                nargs = None
        elif spec.defaults is None:
            nargs = len(spec.args)
        else:
            nargs = len(spec.args) - len(spec.defaults), len(spec.args)

        if prefix:
            namespace = self.namespaces[prefix]
            qname = QName(namespace, f'{prefix}:{symbol}')
        else:
            namespace = XPATH_FUNCTIONS_NAMESPACE
            qname = QName(XPATH_FUNCTIONS_NAMESPACE, f'fn:{symbol}')

        function_signatures = get_function_signatures(qname, nargs, sequence_types)

        if qname.expanded_name in self.symbol_table:
            msg = f'function {qname.qname!r} is already registered'
            raise ElementPathValueError(msg)
        elif symbol not in self.symbol_table:
            lookup_name = symbol
        else:
            lookup_name = qname.expanded_name
            token_class = self.symbol_table[symbol]
            if issubclass(token_class, ProxyToken):
                pass
            elif namespace == token_class.namespace:
                msg = f'function {qname.qname!r} is already registered'
                raise ElementPathValueError(msg)
            elif not issubclass(token_class, XPathFunction) \
                    or token_class.label == 'kind test':
                msg = f'{symbol!r} name collides with {token_class!r}'
                raise ElementPathValueError(msg)
            else:
                # Register a new proxy token, moving the already present token class to
                # qualified name.
                self.symbol_table[f'{{{token_class.namespace}}}{symbol}'] = token_class
                class_name = f'{upper_camel_case(qname.local_name)}FunctionProxy'
                kwargs = {
                    'bases': (ProxyToken,),
                    'label': 'function',
                    'lbp': bp,
                    'rbp': bp,
                }
                self._register(symbol, class_name, **kwargs)

        class_name = f'{upper_camel_case(qname.qname)}ExternalFunction'
        kwargs = {
            'lookup_name': lookup_name,
            'bases': (ExternalFunction,),
            'namespace': namespace,
            'label': 'external function',
            'nargs': nargs,
            'lbp': bp,
            'rbp': bp,
            'sequence_types': sequence_types,
            'callback': staticmethod(callback),
        }
        token_class = self._register(symbol, class_name, **kwargs)

        if function_signatures:
            # Register function signatures
            if self.function_signatures is self.__class__.function_signatures:
                self.function_signatures = dict(self.__class__.function_signatures)
            self.function_signatures.update(function_signatures)

        return cast(type[XPathFunction], token_class)

    def is_schema_bound(self) -> bool:
        return self.schema is not None and 'symbol_table' in self.__dict__

    def check_variables(self, values: MutableMapping[str, Any]) -> None:
        if self.variable_types is None:
            return

        for varname, xsd_type in self.variable_types.items():
            if varname not in values:
                raise xpath_error('XPST0008', "missing variable {!r}".format(varname))

        for varname, value in values.items():
            try:
                sequence_type = self.variable_types[varname]
            except KeyError:
                sequence_type = 'item()*' if isinstance(value, list) else 'item()'

            if not match_sequence_type(value, sequence_type, self):
                message = "Unmatched sequence type for variable {!r}".format(varname)
                raise xpath_error('XPDY0050', message)


##
# Remove symbols that have to be redefined for XPath 2.0.
XPath2Parser.unregister(',')
XPath2Parser.unregister('(')
XPath2Parser.unregister('contains')
XPath2Parser.unregister('lang')
XPath2Parser.unregister('id')
XPath2Parser.unregister('substring-before')
XPath2Parser.unregister('substring-after')
XPath2Parser.unregister('starts-with')

###
# Symbols
XPath2Parser.register('?')
XPath2Parser.register('(:')
XPath2Parser.register(':)')

# XPath 2.0 definitions continue into module xpath2_operators



#
# Copyright (c), 2018-2026, SISSA (International School for Advanced Studies).
# All rights reserved.
# This file is distributed under the terms of the MIT License.
# See the file 'LICENSE' in the root directory of the present
# distribution, or http://opensource.org/licenses/MIT.
#
# @author Davide Brunato <brunato@sissa.it>
#
from itertools import zip_longest
from functools import cache
from typing import cast, Any

import elementpath.aliases as ta

from elementpath.datatypes import builtin_atomic_types, builtin_list_types, QName, \
    NumericProxy, AnyAtomicType
from elementpath.exceptions import ElementPathKeyError, ElementPathValueError, xpath_error
from elementpath.namespaces import XSD_NAMESPACE, XSD_ERROR, XSD_DATETIME_STAMP, \
    XSD_NUMERIC, XSD_UNTYPED, XSD_UNTYPED_ATOMIC, get_expanded_name
from elementpath.helpers import collapse_white_spaces, Patterns
from elementpath.xpath_nodes import XPathNode, DocumentNode, ElementNode, AttributeNode
from elementpath.xpath_tokens import XPathToken

XSD_EXTENDED_PREFIX = f'{{{XSD_NAMESPACE}}}'
XSD_10_UNSUPPORTED = frozenset(
    ('xs:error', 'xs:dateTimeStamp', XSD_ERROR, XSD_DATETIME_STAMP)
)
COMMON_SEQUENCE_TYPES = frozenset((
    'xs:anyType', 'xs:anySimpleType', 'xs:numeric', 'xs:untyped', 'attribute()',
    'attribute(*)', 'element()', 'element(*)', 'text()', 'document-node()',
    'comment()', 'processing-instruction()', 'item()', 'node()', 'numeric'
))
XSD11_ONLY_TYPES = frozenset(
    (XSD_ERROR, XSD_DATETIME_STAMP, 'xs:error', 'xs:dateTimeStamp')
)


def get_function_signatures(qname: QName,
                            nargs: ta.NargsType,
                            sequence_types: tuple[str, ...] = ()) -> dict[tuple[QName, int], str]:
    """
    Returns the signatures for a function.

    :param qname: A QName instance that represents the FQDN of the function.
    :param nargs: The number of arguments that the function takes, could the `None`, \
    a non-negative integer or a couple non-negative integers or a non-negative integer \
    followed by `None`.
    :param sequence_types: A sequence of sequence type specifications, must match \
    the number of arguments that the function takes plus the return type.
    """
    function_signatures: dict[tuple[QName, int], str] = {}
    if not sequence_types:
        return function_signatures

    if any(not is_sequence_type(st) for st in sequence_types):
        msg = "Error in provided sequence types: {!r}"
        raise ElementPathValueError(msg.format(sequence_types))
    elif nargs is None:
        if len(sequence_types) != 1:
            raise ElementPathValueError("Mismatched number of sequence types provided")
        function_signatures[(qname, 0)] = f'function() as {sequence_types[0]}'
    elif isinstance(nargs, int):
        if len(sequence_types) != nargs + 1:
            raise ElementPathValueError("Mismatched number of sequence types provided")
        function_signatures[(qname, nargs)] = 'function({}) as {}'.format(
            ', '.join(sequence_types[:-1]), sequence_types[-1]
        )
    elif nargs[1] is None:
        if len(sequence_types) != nargs[0] + 1:
            raise ElementPathValueError("Mismatched number of sequence types provided")
        function_signatures[(qname, nargs[0])] = 'function({}, ...) as {}'.format(
            ', '.join(sequence_types[:-1]), sequence_types[-1]
        )
    else:
        if len(sequence_types) != nargs[1] + 1:
            raise ElementPathValueError("Mismatched number of sequence types provided")
        for arity in range(nargs[0], nargs[1] + 1):
            function_signatures[(qname, arity)] = 'function({}) as {}'.format(
                ', '.join(sequence_types[:arity]), sequence_types[-1]
            )

    return function_signatures


###
# Sequence type checking
@cache
def normalize_sequence_type(sequence_type: str) -> str:
    sequence_type = collapse_white_spaces(sequence_type)
    sequence_type = Patterns.sequence_type.sub(r'\1', sequence_type)
    return sequence_type.replace(',', ', ').replace(')as', ') as')


@cache
def is_sequence_type_restriction(st1: str, st2: str) -> bool:
    """Returns `True` if st2 is a restriction of st1."""
    st1, st2 = normalize_sequence_type(st1), normalize_sequence_type(st2)

    if not st1 or st1[0] == '{' or not st2 or st2[0] == '{':
        return False
    elif st2 in ('empty-sequence()', 'none') and \
            (st1 in ('empty-sequence()', 'none') or st1.endswith(('?', '*'))):
        return True

    # check occurrences
    if st1[-1] not in '?+*':
        if st2[-1] in '+*':
            return False
        elif st2[-1] == '?':
            st2 = st2[:-1]

    elif st1[-1] == '+':
        st1 = st1[:-1]
        if st2[-1] in '?*':
            return False
        elif st2[-1] == '+':
            st2 = st2[:-1]

    elif st1[-1] == '*':
        st1 = st1[:-1]
        if st2[-1] in '?+':
            return False
        elif st2[-1] == '*':
            st2 = st2[:-1]

    else:
        st1 = st1[:-1]
        if st2[-1] in '+*':
            return False
        elif st2[-1] == '?':
            st2 = st2[:-1]

    if st1 == st2:
        return True
    elif st1 == 'item()':
        return True
    elif st2 == 'item()':
        return False
    elif st1 == 'node()':
        return st2.startswith(('element(', 'attribute(', 'comment(', 'text(',
                               'processing-instruction(', 'document(', 'namespace('))
    elif st2 == 'node()':
        return False
    elif st2 in builtin_atomic_types:
        if st1 not in builtin_atomic_types:
            return False
        return issubclass(builtin_atomic_types[st2], builtin_atomic_types[st1])
    elif st2 in builtin_list_types:
        if st1 in ('xs:anyType', 'xs:anySimpleType'):
            return True
        elif st1 not in builtin_list_types:
            return False
        return issubclass(builtin_list_types[st2], builtin_list_types[st1])
    elif not st1.startswith('function('):
        return False

    if st1 == 'function(*)':
        return st2.startswith('function(')

    parts1 = st1[9:].partition(') as ')
    parts2 = st2[9:].partition(') as ')

    for st1, st2 in zip_longest(parts1[0].split(', '), parts2[0].split(', ')):
        if st1 is None or st2 is None:
            return False
        if not is_sequence_type_restriction(st2, st1):
            return False
    else:
        if not is_sequence_type_restriction(parts1[2], parts2[2]):
            return False
        return True


def is_instance(obj: Any, type_qname: str, parser: ta.XPathParserType | None = None) -> bool:
    """Checks an instance against an XSD type."""
    if isinstance(obj, list) and len(obj) == 1:
        return is_instance(obj[0], type_qname, parser)
    elif type_qname in builtin_atomic_types:
        if type_qname in XSD11_ONLY_TYPES:
            if parser is not None and parser.xsd_version == '1.0':
                return False
        return isinstance(obj, builtin_atomic_types[type_qname])
    elif type_qname in builtin_list_types:
        if isinstance(obj, AnyAtomicType):
            raise xpath_error(
                'XPST0051', 'an atomic value cannot be tested as an instance of list types')
        return isinstance(obj, builtin_list_types[type_qname])
    elif type_qname in (XSD_NUMERIC, 'xs:numeric', 'numeric'):
        return isinstance(obj, NumericProxy)

    elif parser is not None and parser.schema is not None:
        type_qname = get_expanded_name(type_qname, parser.namespaces)
        try:
            return parser.schema.is_instance(obj, type_qname)
        except KeyError:
            pass

    raise ElementPathKeyError("unknown type %r" % type_qname)


def is_sequence_type(value: str, parser: ta.XPathParserType | None = None) -> bool:
    """Checks if a string is a sequence type specification."""

    @cache
    def is_st(st: str) -> bool:
        if not st or st[0] == '{':
            return False
        elif st == 'empty-sequence()' or st == 'none':
            return True
        elif st[-1] in ('*', '+', '?'):
            st = st[:-1]

        if st in COMMON_SEQUENCE_TYPES:
            return True
        elif st in builtin_atomic_types:
            if st in ('xs:dateTimeStamp', 'xs:error'):
                return getattr(parser, 'xsd_version', '1.1') != '1.0'
            return True
        elif st in builtin_list_types:
            return True
        elif st.startswith(('map(', 'array(')):
            if parser and parser.version < '3.1' or not st.endswith(')'):
                return False

            if st in ('map(*)', 'array(*)'):
                return True

            if st.startswith('map('):
                key_type, _, value_type = st[4:-1].partition(', ')
                return key_type.startswith('xs:') and \
                    not key_type.endswith(('+', '*')) and \
                    is_st(key_type) and \
                    is_st(key_type)
            else:
                return is_st(st[6:-1])

        elif st.startswith('element(') and st.endswith(')'):
            if ',' not in st:
                return Patterns.extended_qname.match(st[8:-1]) is not None

            try:
                arg1, arg2 = st[8:-1].split(', ')
            except ValueError:
                return False
            else:
                return (arg1 == '*' or Patterns.extended_qname.match(arg1) is not None) \
                       and Patterns.extended_qname.match(arg2) is not None

        elif st.startswith('document-node(') and st.endswith(')'):
            if not st.startswith('document-node(element('):
                return False
            return is_st(st[14:-1])

        elif st.startswith('function('):
            if parser and parser.version < '3.0':
                return False
            elif st == 'function(*)':
                return True
            elif ' as ' in st:
                pass
            elif not st.endswith(')'):
                return False
            else:
                return is_st(st[9:-1])

            st, return_type = st.rsplit(' as ', 1)
            if not is_st(return_type):
                return False
            elif st == 'function()':
                return True

            st = st[9:-1]
            if st.endswith(', ...'):
                st = st[:-5]

            if 'function(' not in st:
                return all(is_st(x) for x in st.split(', '))
            elif st.startswith('function(*)') and 'function(' not in st[11:]:
                return all(is_st(x) for x in st.split(', '))

            # Cover only if function() spec is the last argument
            k = st.index('function(')
            if not is_st(st[k:]):
                return False
            return all(is_st(x) for x in st[:k].split(', ') if x)

        elif QName.pattern.match(st) is None:
            return False

        if parser is None:
            return False

        try:
            is_instance(None, st, parser)
        except (KeyError, ValueError):
            return False
        else:
            return True

    return is_st(normalize_sequence_type(value))


def match_sequence_type(value: Any,
                        sequence_type: str,
                        parser: ta.XPathParserType | None = None,
                        strict: bool = True) -> bool:
    """
    Checks a value instance against a sequence type.

    :param value: the instance to check.
    :param sequence_type: a string containing the sequence type spec.
    :param parser: an optional parser instance for type checking.
    :param strict: if `False` match xs:anyURI with strings.
    """
    def match_st(v: Any, st: str, occurrence: str | None = None) -> bool:
        if st[-1] in ('*', '+', '?') and ') as ' not in st:
            return match_st(v, st[:-1], st[-1])
        elif v is None or v == []:
            return st in ('empty-sequence()', 'none') or occurrence in ('?', '*')
        elif st in ('empty-sequence()', 'none'):
            return False
        elif isinstance(v, list):
            if len(v) == 1:
                return match_st(v[0], st)
            elif occurrence is None or occurrence == '?':
                return False
            else:
                return all(match_st(x, st) for x in v)
        elif st == 'item()':
            return isinstance(v, (XPathNode, AnyAtomicType,
                                  XPathToken.registry.function_token))
        elif st == 'numeric' or st == 'xs:numeric':
            return isinstance(v, NumericProxy)
        elif st.startswith('function('):
            if not isinstance(v, XPathToken.registry.function_token):
                return False
            return v.match_function_test(st)

        elif st.startswith('array('):
            if not isinstance(v, XPathToken.registry.array_token):
                return False
            if st == 'array(*)':
                return True

            item_st = st[6:-1]
            assert isinstance(v, XPathToken.registry.array_token)
            return all(match_st(x, item_st) for x in v.items())

        elif st.startswith('map('):
            if not isinstance(v, XPathToken.registry.map_token):
                return False
            if st == 'map(*)':
                return True

            key_st, _, value_st = st[4:-1].partition(', ')
            if key_st.endswith(('+', '*')):
                raise xpath_error('XPST0003', 'no multiples occurs for a map key')

            return all(match_st(k, key_st) and match_st(v, value_st) for k, v in v.items())

        if isinstance(v, XPathNode):
            node_kind = v.node_kind
        elif '(' in st:
            return False
        elif not strict and st == 'xs:anyURI' and isinstance(v, str):
            return True
        else:
            try:
                return is_instance(v, st, parser)
            except (KeyError, ValueError):
                raise xpath_error('XPST0051')

        if st == 'node()':
            return True
        elif not st.startswith(node_kind) or not st.endswith(')'):
            return False
        elif st == f'{node_kind}()':
            return True
        elif node_kind == 'document':
            element_test = st[14:-1]
            if not element_test:
                return True
            document = cast(DocumentNode, v)
            return any(
                match_st(e, element_test) for e in document if isinstance(e, ElementNode)
            )
        elif node_kind not in ('element', 'attribute'):
            return False

        _, params = st[:-1].split('(')
        if ', ' not in st:
            name = params
        else:
            name, type_name = params.rsplit(', ', 1)
            if type_name.endswith('?'):
                type_name = type_name[:-1]
            elif isinstance(v, ElementNode) and v.nilled:
                return False

            if type_name == 'xs:untyped':
                if isinstance(v, AttributeNode) and v.type_name != XSD_UNTYPED_ATOMIC:
                    return False
                if isinstance(v, ElementNode) and v.type_name != XSD_UNTYPED:
                    return False
            else:
                try:
                    if not is_instance(v.typed_value, type_name, parser):
                        return False
                except (KeyError, ValueError):
                    raise xpath_error('XPST0051')

        if name == '*':
            return True
        elif parser is None:
            return v.name == name
        else:
            try:
                return v.name == get_expanded_name(name, parser.namespaces)
            except (KeyError, ValueError):
                return False

    return match_st(value, normalize_sequence_type(sequence_type))
